@@ -25,8 +25,8 @@ import inflect
 import pint
 from albert import *
 
-md_iid = "4.0"
-md_version = "1.8.1"
+md_iid = "5.0"
+md_version = "1.8.2"
 md_name = "Unit Converter"
 md_description = "Convert between units"
 md_license = "MIT"
@@ -361,13 +361,12 @@ class Plugin(PluginInstance, GlobalQueryHandler):
     def synopsis(self, query):
         return "<amount> <from_unit> to <to_unit>"
 
-    def handleTriggerQuery(self, query: Query) -> None:
-        if query_string := query.string.strip():
-            items = self.match_query(query_string)
-            query.add(items)
+    def items(self, ctx):
+        if query_string := ctx.query.strip():
+            yield self.match_query(query_string)
 
-    def handleGlobalQuery(self, query):
-        return [RankItem(item=item, score=1) for item in self.match_query(query.string.strip())]
+    def rankItems(self, ctx):
+        return [RankItem(item=item, score=1) for item in self.match_query(ctx.query.strip())]
 
     def match_query(self, query_string: str):
         match = self.unit_convert_regex.fullmatch(query_string)
@@ -403,7 +402,7 @@ class Plugin(PluginInstance, GlobalQueryHandler):
             icon_path = Path(__file__).parent / "icons" / "unit_converter.svg"
         return StandardItem(
             id=str(icon_path),
-            icon_factory=lambda: makeImageIcon(icon_path),
+            icon_factory=lambda: Icon.image(icon_path),
             text=text,
             subtext=subtext,
             actions=[
